@@ -3,26 +3,25 @@
 ;; Copyright (C) 1998 Andy Piper.
 ;; Copyright (C) 1990, 1997 Free Software Foundation, Inc.
 ;; Copyright (C) 1995 Sun Microsystems.
+;; Copyright (C) 2002 Ben Wing.
 
 ;; Maintainer: XEmacs Development Team
 ;; Keywords: extensions, dumped
 
 ;; This file is part of XEmacs.
 
-;; XEmacs is free software; you can redistribute it and/or modify it
-;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
-;; any later version.
+;; XEmacs is free software: you can redistribute it and/or modify it
+;; under the terms of the GNU General Public License as published by the
+;; Free Software Foundation, either version 3 of the License, or (at your
+;; option) any later version.
 
-;; XEmacs is distributed in the hope that it will be useful, but
-;; WITHOUT ANY WARRANTY; without even the implied warranty of
-;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-;; General Public License for more details.
+;; XEmacs is distributed in the hope that it will be useful, but WITHOUT
+;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+;; for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with XEmacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with XEmacs.  If not, see <http://www.gnu.org/licenses/>.
 
 ;;; Synched up with: Not in FSF
 
@@ -32,13 +31,24 @@
 
 ;;; Code:
 
-(defvar selected-text-type
-  (if (featurep 'mule) '(COMPOUND_TEXT STRING) 'STRING)
-  "The type atom used to obtain selections from the X server.
-Can be either a valid X selection data type, or a list of such types.
-COMPOUND_TEXT and STRING are the most commonly used data types.
-If a list is provided, the types are tried in sequence until
-there is a successful conversion.")
+(globally-declare-fboundp
+ (unless (featurep 'window-system)
+   '(default-mouse-track-next-move-rect
+     mouse-track-activate-rectangular-selection)))
+
+;; We prefer UTF8_STRING to COMPOUND_TEXT because, even though the latter
+;; gives us more information when taking data from other XEmacs invocations,
+;; Mozilla will happily give us broken COMPOUND_TEXT where a non-broken
+;; UTF8_STRING is available.
+(defvar selection-preferred-types
+  `(UTF8_STRING text/plain\;charset=utf-8 COMPOUND_TEXT STRING text/plain
+    ,@(mapcan #'(lambda (format)
+                  (and (featurep format)
+                       (list (intern (format "image/%s" format)))))
+              '(png gif jpeg tiff xpm xbm)))
+  "An ordered list of X11 type atoms for selections we want to receive.
+We prefer UTF8_STRING over COMPOUND_TEXT, and COMPOUND_TEXT over STRING.  We
+also accept several image types.")
 
 (defvar selection-sets-clipboard nil
   "Controls the selection's relationship to the clipboard.
@@ -46,13 +56,19 @@ When non-nil, any operation that sets the primary selection will also
 set the clipboard.")
 
 (defun copy-primary-selection ()
-  "Copy the selection to the Clipboard and the kill ring."
+  "Copy the selection to the Clipboard and the kill ring.
+This is similar to the command \\[kill-ring-save] except that it will
+save to the Clipboard even if that command doesn't, and it handles rectangles
+properly."
   (interactive)
   (and (console-on-window-system-p)
        (cut-copy-clear-internal 'copy)))
 
 (defun kill-primary-selection ()
-  "Copy the selection to the Clipboard and the kill ring, then delete it."
+  "Copy the selection to the Clipboard and the kill ring, then delete it.
+This is similar to the command \\[kill-region] except that it will
+save to the Clipboard even if that command doesn't, and it handles rectangles
+properly."
   (interactive "*")
   (and (console-on-window-system-p)
        (cut-copy-clear-internal 'cut)))
@@ -90,34 +106,57 @@ This will do nothing under anything other than X.")
 
 (defun get-selection-no-error (&optional type data-type)
   "Return the value of a window-system selection.
-The argument TYPE (default `PRIMARY') says which selection,
-and the argument DATA-TYPE (default `STRING', or `COMPOUND_TEXT' under Mule)
-says how to convert the data. Returns NIL if there is no selection."
+The argument TYPE (default `PRIMARY') says which selection, and the argument
+DATA-TYPE (defaulting to the value of `selection-preferred-types'), says how
+to convert the data. Returns NIL if there is no selection."
   (condition-case nil (get-selection type data-type) (t nil)))
 
 (defun get-selection (&optional type data-type)
   "Return the value of a window-system selection.
-The argument TYPE (default `PRIMARY') says which selection,
-and the argument DATA-TYPE (default `STRING', or `COMPOUND_TEXT' under Mule)
-says how to convert the data. If there is no selection an error is signalled.
-Not suitable in a `interprogram-paste-function', q.v."
+The argument TYPE (default `PRIMARY') says which selection, and the argument
+DATA-TYPE (defaulting to the value of, and compatible with,
+`selection-preferred-types') says how to convert the data. If
+there is no selection an error is signalled.  Not suitable in a
+`interprogram-paste-function', q.v."
   (or type (setq type 'PRIMARY))
-  (or data-type (setq data-type selected-text-type))
+  (or data-type (setq data-type selection-preferred-types))
   (if (consp data-type)
-      (condition-case err
-	  (get-selection-internal type (car data-type))
-	(selection-conversion-error
-	 (if (cdr data-type)
-	     (get-selection type (cdr data-type))
-	   (signal (car err) (cdr err)))))
+      ;; TARGETS is a vector; we want a list so we can memq --> append it to
+      ;; nil.
+      (let ((targets (append (get-selection-internal type 'TARGETS) nil))
+	    res)
+	(catch 'converted
+	  (if targets
+	      (dolist (current-preference data-type)
+		(condition-case nil
+		    (if (and (memq current-preference targets)
+			     (setq res (get-selection-internal
+					type current-preference)))
+			(throw 'converted res))
+		  (selection-conversion-error
+		   nil))))
+	  ;; The source app didn't offer us anything compatible in TARGETS,
+	  ;; or they're not negotiating at all. (That is, we're probably not
+	  ;; on X11.) Try to convert to the types specified by our caller,
+	  ;; and throw an error if the last one of those fails.
+	  (while data-type
+	    (condition-case err
+		(progn
+		  (setq res (get-selection-internal type (car data-type)))
+		  (if res (throw 'converted res) 
+		    (signal 'selection-conversion-error nil)))
+	      (selection-conversion-error
+	       (if (cdr data-type)
+		   (setq data-type (cdr data-type))
+		 (signal (car err) (cdr err))))))))
     (get-selection-internal type data-type)))
 
 (defun get-selection-foreign (&optional type data-type)
   "Return the value of a window-system selection, or nil if XEmacs owns it.
-The argument TYPE (default `PRIMARY') says which selection,
-and the argument DATA-TYPE (default `STRING', or `COMPOUND_TEXT' under Mule)
-says how to convert the data. If there is no selection an error is signalled.
-See `interprogram-paste-function' for more information."
+The argument TYPE (default `PRIMARY') says which selection, and the argument
+DATA-TYPE (defaulting to the value of `selection-preferred-types' which see)
+says how to convert the data. If there is no selection an error is
+signalled.  See `interprogram-paste-function' for more information."
   (unless (selection-owner-p type)
     (get-selection type data-type)))
 
@@ -130,7 +169,7 @@ and DATA specifies the contents.  DATA may be any lisp data type
 that can be converted using the function corresponding to DATA-TYPE
 in `select-converter-alist'---strings are the usual choice, but
 other types may be permissible depending on the DATA-TYPE parameter
-(if DATA-TYPE is not supplied, the default behavior is window
+\(if DATA-TYPE is not supplied, the default behavior is window
 system specific, but strings are always accepted).
 HOW-TO-ADD may be any of the following:
 
@@ -139,7 +178,7 @@ HOW-TO-ADD may be any of the following:
   'append or t        -- append data to existing DATA-TYPE data.
 
 DATA-TYPE is the window-system specific data type identifier
-(see `register-selection-data-type' for more information).
+\(see `register-selection-data-type' for more information).
 
 The selection may also be a cons of two markers pointing to the same buffer,
 or an overlay.  In these cases, the selection is considered to be the text
@@ -180,7 +219,7 @@ Interactively, the text of the region is used as the selection value."
   ;; when behaving as the latter, we better not set it, or we will
   ;; cause unwanted sticky-region behavior in kill-region and friends.
   (if (interactive-p)
-  (setq zmacs-region-stays t))
+      (setq zmacs-region-stays t))
   data)
 
 (defun dehilight-selection (selection)
@@ -189,7 +228,7 @@ Interactively, the text of the region is used as the selection value."
 	 (if primary-selection-extent
 	     (let ((inhibit-quit t))
 	       (if (consp primary-selection-extent)
-		   (mapcar 'delete-extent primary-selection-extent)
+		   (mapc 'delete-extent primary-selection-extent)
 		 (delete-extent primary-selection-extent))
 	       (setq primary-selection-extent nil)))
 	 (if zmacs-regions (zmacs-deactivate-region)))
@@ -197,7 +236,7 @@ Interactively, the text of the region is used as the selection value."
 	 (if secondary-selection-extent
 	     (let ((inhibit-quit t))
 	       (if (consp secondary-selection-extent)
-		   (mapcar 'delete-extent secondary-selection-extent)
+		   (mapc 'delete-extent secondary-selection-extent)
 		 (delete-extent secondary-selection-extent))
 	       (setq secondary-selection-extent nil)))))
   nil)
@@ -215,12 +254,9 @@ With an argument, discard the secondary selection instead of the
 primary selection."
   (disown-selection-internal (if secondary-p 'SECONDARY 'PRIMARY))
   (when (and selection-sets-clipboard
-	     (or (not secondary-p)
-		 (eq secondary-p 'PRIMARY)
-		 (eq secondary-p 'CLIPBOARD)))
+             (memq secondary-p '(nil PRIMARY CLIPBOARD)))
     (disown-selection-internal 'CLIPBOARD)))
 
-;; from x-init.el
 ;; selections and active regions
 
 ;; If and only if zmacs-regions is true:
@@ -235,10 +271,12 @@ primary selection."
 ;; application asserts the selection.  This is probably not a big deal.
 
 (defun activate-region-as-selection ()
-  (if (marker-buffer (mark-marker t))
-      (own-selection (cons (point-marker t) (mark-marker t)))))
+  (cond ((and-fboundp 'mouse-track-rectangle-p
+           (mouse-track-rectangle-p
+            (mouse-track-activate-rectangular-selection))))
+	((marker-buffer (mark-marker t))
+	 (own-selection (cons (point-marker t) (mark-marker t))))))
 
-; moved from x-select.el
 (defvar primary-selection-extent nil
   "The extent of the primary selection; don't use this.")
 
@@ -249,6 +287,8 @@ primary selection."
   ;; Given a selection, this makes an extent in the buffer which holds that
   ;; selection, for highlighting purposes.  If the selection isn't associated
   ;; with a buffer, this does nothing.
+  ;; 
+  ;; Something similar needs to be hooked into the rectangle functions. 
   (let ((buffer nil)
 	(valid (and (extentp previous-extent)
 		    (extent-object previous-extent)
@@ -279,7 +319,7 @@ primary selection."
 	nil
       (condition-case ()
 	  (if (listp previous-extent)
-	      (mapcar 'delete-extent previous-extent)
+	      (mapc 'delete-extent previous-extent)
 	    (delete-extent previous-extent))
 	(error nil)))
 
@@ -303,13 +343,13 @@ primary selection."
 	(set-extent-property previous-extent 'end-open nil)
 
 	(cond
-	 (mouse-track-rectangle-p
-	  (setq previous-extent (list previous-extent))
-	  (default-mouse-track-next-move-rect start end previous-extent)
-	  ))
+	 ((and-fboundp 'mouse-track-rectangle-p 
+            (mouse-track-rectangle-p
+             (setq previous-extent (list previous-extent))
+             (default-mouse-track-next-move-rect start end previous-extent)
+             ))))
 	previous-extent))))
 
-;; moved from x-select.el
 (defun valid-simple-selection-p (data)
   "An obsolete function that tests whether something was a valid simple
 selection using the old XEmacs selection support. You shouldn't use this
@@ -334,7 +374,7 @@ any more, because just about anything could be a valid selection now."
 	   (buffer-live-p (marker-buffer (cdr data))))))
 
 (defun cut-copy-clear-internal (mode)
-  (or (memq mode '(cut copy clear)) (error "unkown mode %S" mode))
+  (or (memq mode '(cut copy clear)) (error "unknown mode %S" mode))
   (or (selection-owner-p)
       (error "XEmacs does not own the primary selection"))
   (setq last-command nil)
@@ -357,8 +397,7 @@ any more, because just about anything could be a valid selection now."
       (cond ((memq mode '(cut copy))
 	     (if rect-p
 		 (progn
-		   ;; why is killed-rectangle free?  Is it used somewhere?
-		   ;; should it be defvarred?
+		   ;; killed-rectangle is defvarred in rect.el
 		   (setq killed-rectangle (extract-rectangle s e))
 		   (kill-new (mapconcat #'identity killed-rectangle "\n")))
 	       (copy-region-as-kill s e))
@@ -367,6 +406,7 @@ any more, because just about anything could be a valid selection now."
 	     ;; some other way, but owning the clipboard twice in that case
 	     ;; wouldn't actually hurt anything.
 	     (or (and (consp kill-hooks) (memq 'own-clipboard kill-hooks))
+		 (eq 'own-clipboard interprogram-cut-function)
 		 (own-clipboard (car kill-ring)))))
       (cond ((memq mode '(cut clear))
 	     (if rect-p
@@ -385,40 +425,40 @@ any more, because just about anything could be a valid selection now."
 for the specified SELECTION. Return nil if this is impossible, or a
 suitable internal representation otherwise."
   (when value
-    (let ((handler-fn (cdr (assq type selection-converter-in-alist))))
-      (when handler-fn
-	(apply handler-fn (list selection type value))))))
+    (let ((handler (cdr (assq type selection-converter-in-alist))))
+      (if handler
+          (funcall handler selection type value)
+        ;; Else, the value is already in Lisp, pass it through.
+        value))))
 
 (defun select-convert-out (selection type value)
   "Attempt to convert the specified internal VALUE for the specified DATA-TYPE
 and SELECTION. Return nil if this is impossible, or a suitable external
 representation otherwise."
-  (when value
-    (let ((handler-fn (cdr (assq type selection-converter-out-alist))))
-      (when handler-fn
-	(apply handler-fn (list selection type value))))))
+  (and value
+       (funcall (or (cdr (assq type selection-converter-out-alist)) #'ignore)
+                selection type value)))
 
 (defun select-coerce (selection type value)
   "Attempt to convert the specified internal VALUE to a representation
 suitable for return from `get-selection' in the specified DATA-TYPE. Return
 nil if this is impossible, or a suitable representation otherwise."
-  (when value
-    (let ((handler-fn (cdr (assq type selection-coercion-alist))))
-      (when handler-fn
-	(apply handler-fn (list selection type value))))))
+  (and value
+       (funcall (or (cdr (assq type selection-coercion-alist)) #'ignore)
+                selection type value)))
 
 ;; The rest of the functions on this "page" are conversion handlers,
 ;; append handlers and buffer-kill handlers.
 (defun select-convert-to-text (selection type value)
   (cond ((stringp value)
-	 value)
+	 (substring-no-properties value))
 	((extentp value)
 	 (save-excursion
 	   (set-buffer (extent-object value))
 	   (save-restriction
 	     (widen)
-	     (buffer-substring (extent-start-position value)
-			       (extent-end-position value)))))
+	     (buffer-substring-no-properties (extent-start-position value)
+                                             (extent-end-position value)))))
 	((and (consp value)
 	      (markerp (car value))
 	      (markerp (cdr value)))
@@ -431,16 +471,47 @@ nil if this is impossible, or a suitable representation otherwise."
 			   (error "selection is in a killed buffer")))
 	   (save-restriction
 	     (widen)
-	     (buffer-substring (car value) (cdr value)))))
+	     (buffer-substring-no-properties (car value) (cdr value)))))
+	(t nil)))
+
+(defun select-convert-to-timestamp (selection type value)
+  (let ((ts (get-xemacs-selection-timestamp selection)))
+    (if ts (cons 'TIMESTAMP ts))))
+
+(defun select-convert-to-utf-8-text (selection type value)
+  (cond ((stringp value)
+	 (cons 'UTF8_STRING (encode-coding-string
+                             (substring-no-properties value) 'utf-8)))
+	((extentp value)
+	 (save-excursion
+	   (set-buffer (extent-object value))
+	   (save-restriction
+	     (widen)
+	     (cons 'UTF8_STRING 
+		   (encode-coding-string (buffer-substring-no-properties
+                                          (extent-start-position value)
+                                          (extent-end-position value))
+                                         'utf-8)))))
+	((and (consp value)
+	      (markerp (car value))
+	      (markerp (cdr value)))
+	 (or (eq (marker-buffer (car value)) (marker-buffer (cdr value)))
+	     (signal 'error
+		     (list "markers must be in the same buffer"
+			   (car value) (cdr value))))
+	 (save-excursion
+	   (set-buffer (or (marker-buffer (car value))
+			   (error "selection is in a killed buffer")))
+	   (save-restriction
+	     (widen)
+	     (cons 'UTF8_STRING (encode-coding-string 
+				 (buffer-substring-no-properties
+                                  (car value) (cdr value)) 'utf-8)))))
 	(t nil)))
 
 (defun select-coerce-to-text (selection type value)
   (select-convert-to-text selection type value))
 
-(defun select-convert-from-text (selection type value)
-  (when (stringp value)
-    value))
-
 (defun select-convert-to-string (selection type value)
   (let ((outval (select-convert-to-text selection type value)))
     ;; force the string to be not in Compound Text format. This grubby
@@ -472,19 +543,9 @@ nil if this is impossible, or a suitable representation otherwise."
 	(cons (ash value -16) (logand value 65535))
       nil)))
 
-(defun select-convert-from-length (selection type value)
-  (select-convert-to-length selection type value))
-
 (defun select-convert-to-targets (selection type value)
   ;; return a vector of atoms, but remove duplicates first.
-  (let* ((all (cons 'TIMESTAMP (mapcar 'car selection-converter-alist)))
-	 (rest all))
-    (while rest
-      (cond ((memq (car rest) (cdr rest))
-	     (setcdr rest (delq (car rest) (cdr rest))))
-	    (t
-	     (setq rest (cdr rest)))))
-    (apply 'vector all)))
+  (delete-duplicates (map 'vector #'car selection-converter-out-alist)))
 
 (defun select-convert-to-delete (selection type value)
   (disown-selection-internal selection)
@@ -504,10 +565,6 @@ nil if this is impossible, or a suitable representation otherwise."
 			       (error "selection is in a killed buffer"))))
 	(t nil)))
 
-(defun select-convert-from-filename (selection type value)
-  (when (stringp value)
-    value))
-
 (defun select-convert-to-charpos (selection type value)
   (let (a b tmp)
     (cond ((cond ((extentp value)
@@ -533,8 +590,8 @@ nil if this is impossible, or a suitable representation otherwise."
 		 ((and (consp value)
 		       (markerp (car value))
 		       (markerp (cdr value)))
-		  (setq a (marker-position (car value))
-			b (marker-position (cdr value))
+		  (setq a (car value)
+			b (cdr value)
 			buf (marker-buffer (car value)))))
 	   (save-excursion
 	     (set-buffer buf)
@@ -585,8 +642,8 @@ nil if this is impossible, or a suitable representation otherwise."
 		 ((and (consp value)
 		       (markerp (car value))
 		       (markerp (cdr value)))
-		  (setq a (marker-position (car value))
-			b (marker-position (cdr value))
+		  (setq a (car value)
+			b (cdr value)
 			buf (or (marker-buffer (car value))
 				(error "selection is in a killed buffer"))
 			file-name (buffer-file-name buf))))
@@ -662,34 +719,103 @@ nil if this is impossible, or a suitable representation otherwise."
 
 	((listp value)			; list
 	 (if (cdr value)
-	     (mapcar '(lambda (x)
-			(select-convert-from-integer selection type x))
+	     (mapcar #'(lambda (x)
+                         (select-convert-from-integer selection type x))
 		     value)
 	   (select-convert-from-integer selection type (car value))))
 
 	((vectorp value)		; vector
 	 (if (eq (length value) 1)
 	     (select-convert-from-integer selection type (aref value 0))
-	   (mapvector '(lambda (x)
-			(select-convert-from-integer selection type x))
-		     value)))
+	   (mapvector #'(lambda (x)
+                          (select-convert-from-integer selection type x))
+                      value)))
 
 	(t nil)
 	))
 
+(defun select-convert-from-ip-address (selection type value)
+  (if (and (stringp value)
+           (= (length value) 4))
+      (format "%d.%d.%d.%d"
+              (aref value 0) (aref value 1) (aref value 2) (aref value 3))))
+
 (defun select-convert-to-atom (selection type value)
   (and (symbolp value) value))
 
+(defun select-convert-from-utf-8-text (selection type value)
+  (decode-coding-string value 'utf-8))
+
+(defun select-convert-from-utf-16-le-text (selection type value)
+  (decode-coding-string value 'utf-16-le))
+
+(defun select-convert-from-undecided-text (selection type value)
+  (decode-coding-string value 'undecided))
+
+;; Image conversion. 
+(defun select-convert-from-image-data (image-type value)
+  "Take an image type specification--one of the image types this XEmacs
+supports--and some data in that format, return a space, with a glyph
+corresponding to that data as an end-glyph extent property of that space. "
+  (let* ((str (make-string 1 ?\ ))
+	 (extent (make-extent 0 1 str))
+	 (glyph (make-glyph (vector image-type :data value))))
+    (when glyph
+      (set-extent-property extent 'invisible t)
+      (set-extent-property extent 'start-open t)
+      (set-extent-property extent 'end-open t)
+      (set-extent-property extent 'duplicable t)
+      (set-extent-property extent 'atomic t)
+      (set-extent-end-glyph extent glyph)
+      str)))
+
+(macrolet
+    ((create-image-functions (&rest formats)
+       (cons
+        'progn
+        (mapcar
+         #'(lambda (format)
+             `(if (featurep ',format)
+                  (defalias (intern (concat "select-convert-from-image/"
+                                            ,(symbol-name format)))
+                    #'(lambda (selection type value)
+                       (select-convert-from-image-data ',format
+                                                       value))))) formats))))
+  (create-image-functions gif jpeg png tiff xpm xbm))
+
 ;;; CF_xxx conversions
 (defun select-convert-from-cf-text (selection type value)
-  (replace-in-string (if (string-match "\0" value)
-			 (substring value 0 (match-beginning 0))
-		       value)
-		     "\\(\r\n\\|\n\r\\)" "\n" t))
+  (if (find-coding-system 'mswindows-multibyte)
+      (let* ((value (decode-coding-string value 'mswindows-multibyte))
+             (position (position ?\x00 value)))
+	(replace-in-string (subseq value 0 position)
+			   "\\(\r\n\\|\n\r\\)" "\n" t))))
+
+(defun select-convert-from-cf-unicodetext (selection type value)
+  (cond
+    ((find-coding-system 'mswindows-unicode)
+     (let* ((value (decode-coding-string value 'mswindows-unicode))
+            (position (position ?\x00 value)))
+       (replace-in-string (subseq value 0 position)
+                          "\\(\r\n\\|\n\r\\)" "\n" t)))
+    ((find-coding-system 'utf-16-le-dos)
+     (let* ((value (decode-coding-string value 'utf-16-le-dos))
+            (position (position ?\x00 value)))
+       (subseq value 0 position)))))
 
 (defun select-convert-to-cf-text (selection type value)
-  (let ((text (select-convert-to-text selection type value)))
-    (concat (replace-in-string text "\n" "\r\n" t) "\0")))
+  (if (find-coding-system 'mswindows-multibyte)
+      (let ((text (select-convert-to-text selection type value)))
+	(encode-coding-string
+	 (concat (replace-in-string text "\n" "\r\n" t) "\0")
+	 'mswindows-multibyte))))
+
+(defun select-convert-to-cf-unicodetext (selection type value)
+  (if (find-coding-system 'mswindows-unicode)
+      (let ((text (select-convert-to-text selection type value)))
+	(encode-coding-string
+	 (concat (replace-in-string text "\n" "\r\n" t) "\0")
+	 'mswindows-unicode))))
 
 ;;; Appenders
 (defun select-append-to-text (selection type value1 value2)
@@ -712,6 +838,15 @@ nil if this is impossible, or a suitable representation otherwise."
 	(select-convert-to-cf-text selection type (concat text1 text2))
       nil)))
 
+(defun select-append-to-cf-unicodetext (selection type value1 value2)
+  (let ((text1 (select-convert-from-cf-unicodetext selection
+						    'CF_UNICODETEXT value1))
+	(text2 (select-convert-from-cf-unicodetext selection
+						    'CF_UNICODETEXT value2)))
+    (if (and text1 text2)
+	(select-convert-to-cf-unicodetext selection type (concat text1 text2))
+      nil)))
+
 (defun select-append-default (selection type value1 value2)
 ;; This appender gets used if the type is "nil" - i.e. default.
 ;; It should probably have more cases implemented than it does - e.g.
@@ -758,7 +893,10 @@ nil if this is impossible, or a suitable representation otherwise."
 
 ;; Types listed in here can be selections of XEmacs
 (setq selection-converter-out-alist
-      '((TEXT . select-convert-to-text)
+      `((TIMESTAMP . select-convert-to-timestamp)
+	(UTF8_STRING . select-convert-to-utf-8-text)
+	(text/plain\;charset=utf-8 . select-convert-to-utf-8-text)
+	(TEXT . select-convert-to-text)
 	(STRING . select-convert-to-string)
 	(COMPOUND_TEXT . select-convert-to-compound-text)
 	(TARGETS . select-convert-to-targets)
@@ -776,27 +914,38 @@ nil if this is impossible, or a suitable representation otherwise."
 	(NAME . select-convert-to-name)
 	(ATOM . select-convert-to-atom)
 	(INTEGER . select-convert-to-integer)
-	(CF_TEXT . select-convert-to-cf-text)
-	))
+	,@(when (find-coding-system 'mswindows-multibyte)
+                '((CF_TEXT . select-convert-to-cf-text)))
+	(CF_UNICODETEXT . select-convert-to-cf-unicodetext)))
 
 ;; Types listed here can be selections foreign to XEmacs
 (setq selection-converter-in-alist
-      '(; Specific types that get handled by generic converters
-	(COMPOUND_TEXT . select-convert-from-text)
-	(SOURCE_LOC . select-convert-from-text)
-	(OWNER_OS . select-convert-from-text)
-	(HOST_NAME . select-convert-from-text)
-	(USER . select-convert-from-text)
-	(CLASS . select-convert-from-text)
-	(NAME . select-convert-from-text)
-	; Generic types
+      `(; Specific types that get handled by generic converters
 	(INTEGER . select-convert-from-integer)
-	(TEXT . select-convert-from-text)
-	(STRING . select-convert-from-text)
-	(LENGTH . select-convert-from-length)
-	(FILE_NAME . select-convert-from-filename)
-	(CF_TEXT . select-convert-from-cf-text)
-	))
+ 	(TIMESTAMP . select-convert-from-integer)
+ 	(LENGTH . select-convert-from-integer)
+ 	(LIST_LENGTH . select-convert-from-integer)
+ 	(CLIENT_WINDOW . select-convert-from-integer)
+ 	(PROCESS . select-convert-from-integer)
+ 	(IP_ADDRESS . select-convert-from-ip-address)
+	;; We go after UTF8_STRING in preference to STRING because Mozilla,
+	;; at least, does bad things with non-Latin-1 Unicode characters in
+	;; STRING.
+ 	(UTF8_STRING . select-convert-from-utf-8-text)
+        (text/plain\;charset=utf-8 . select-convert-from-utf-8-text)
+	,@(when (find-coding-system 'mswindows-multibyte)
+                '((CF_TEXT . select-convert-from-cf-text)))
+	(CF_UNICODETEXT . select-convert-from-cf-unicodetext)
+ 	(text/html . select-convert-from-utf-16-le-text)  ; Mozilla
+ 	(text/_moz_htmlcontext . select-convert-from-utf-16-le-text)
+ 	(text/_moz_htmlinfo . select-convert-from-utf-16-le-text)
+        (text/plain . select-convert-from-undecided-text)
+        ,@(loop
+            for format in '(gif jpeg png tiff xpm xbm)
+            nconc (if (featurep format)
+                      (list (cons (intern (format "image/%s" format))
+                                  (intern (format "select-convert-from-image/%s"
+                                                  format))))))))
 
 ;; Types listed here have special coercion functions that can munge
 ;; other types. This can also be used to add special features - e.g.
@@ -812,26 +961,33 @@ nil if this is impossible, or a suitable representation otherwise."
       '((TEXT . select-coerce-to-text)
 	(STRING . select-coerce-to-text)
 	(COMPOUND_TEXT . select-coerce-to-text)
-	(CF_TEXT . select-coerce-to-text)))
+	(CF_TEXT . select-coerce-to-text)
+	(CF_UNICODETEXT . select-coerce-to-text)
+	))
 
 ;; Types listed here can be appended by own-selection
 (setq selection-appender-alist
-      '((nil . select-append-default)
+      `((nil . select-append-default)
 	(TEXT . select-append-to-text)
 	(STRING . select-append-to-string)
 	(COMPOUND_TEXT . select-append-to-compound-text)
-	(CF_TEXT . select-append-to-cf-text)
+	,@(when (find-coding-system 'mswindows-multibyte)
+                '((CF_TEXT . select-append-to-cf-text)))
+	(CF_UNICODETEXT . select-append-to-cf-unicodetext)
 	))
 
 ;; Types listed here have buffer-kill handlers
 (setq selection-buffer-killed-alist
-      '((nil . select-buffer-killed-default)
+      `((nil . select-buffer-killed-default)
 	(TEXT . select-buffer-killed-text)
 	(STRING . select-buffer-killed-text)
 	(COMPOUND_TEXT . select-buffer-killed-text)
-	(CF_TEXT . select-buffer-killed-text)))
+	,@(when (find-coding-system 'mswindows-multibyte)
+                '((CF_TEXT . select-buffer-killed-text)))
+	(CF_UNICODETEXT . select-buffer-killed-text)))
 
 ;; Lists of types that are coercible (can be converted to other types)
-(setq selection-coercible-types '(TEXT STRING COMPOUND_TEXT))
+(setq selection-coercible-types
+      '(TEXT STRING COMPOUND_TEXT CF_TEXT CF_UNICODETEXT))
 
 ;;; select.el ends here
